"""# Testing
"""
import logging
import typing as tp
from concurrent import futures
from operator import methodcaller
from pathlib import Path
import nbclient
import nbformat
from traitlets import Bool, Int, List, Type, Unicode, default
from .project import ProjectOperator
DEFAULT_IGNORE_PATTERNS = ('.ipynb_checkpoints', '__pycache__', '.*')

class ProjectTester(ProjectOperator):
    """Test the current project notebooks, with various failure strategies."""
    client_class = Type(nbclient.NotebookClient, help='Class for the notebook client').tag(config=True)
    extra_sources = List(Unicode(), help='List of paths to extra sources to be tested').tag(config=True)

    @property
    def extra_paths(self) -> tp.List[Path]:
        return [self.resolve_path(p) for p in self.extra_sources]
    fail_fast = Bool(default_value=True, help='Fail as soon as one client fails').tag(config=True)
    ignore_patterns = List(Unicode(), help='List of patterns to ignore from sources').tag(config=True)

    @default('ignore_patterns')
    def _ignore_patterns_default(self):
        return list(DEFAULT_IGNORE_PATTERNS)
    jobs = Int(allow_none=True, default_value=None, help='number of parallel jobs to run').tag(config=True)

    def _execute_notebook(self, path: Path):
        """Execute the notebook with the given path.
    
    :param path: path to notebook
    """
        nb = nbformat.read(path, as_version=nbformat.NO_CONVERT)
        client = self.client_class(nb, resources={'metadata': {'path': path.parent}}, parent=self)
        return client.execute()

    def _find_notebooks(self, dir_path: Path):
        """Find notebooks given by a particular path.

    If the path is a directory, yield from the result of calling find_notebooks` with
    the directory path.
    If the path is a notebook file path, yield the path directly

    :param path: path to a file or directory
    :param ignore_patterns: set of patterns to ignore during recursion
    :return:
    """
        for path in dir_path.iterdir():
            if any((path.match(p) for p in self.ignore_patterns)):
                continue
            if path.is_dir():
                yield from self._find_notebooks(path)
            elif path.match('*.ipynb'):
                yield path

    def _visit_and_flatten_paths(self, paths: tp.Iterable[str]):
        """Flatten an iterable of directory and file paths into file paths.
    Directories will be visited and any notebook paths that are not ignored will be yielded.
    
    :param paths: iterable of paths
    """
        for p in paths:
            if p.is_dir():
                yield from self._find_notebooks(p)
            else:
                yield p

    def run(self):
        """Run the tracked notebooks in a process pool."""
        sources = [Path(p) for p in [self.packages_path] + self.extra_paths]
        paths = [*self._visit_and_flatten_paths(sources)]
        return_when = futures.FIRST_EXCEPTION if self.fail_fast else futures.ALL_COMPLETED
        with futures.ProcessPoolExecutor(max_workers=self.jobs) as executor:
            tasks = [executor.submit(self._execute_notebook, p) for p in paths]
            (done, not_done) = futures.wait(tasks, return_when=return_when)
            for (path, task) in zip(paths, done):
                task.result()
                self.log.info(f'{path} executed successfully')


"""# Import Hook
"""
import sys
from pathlib import Path
from ..notebook.importer import NotebookImporter
from ..notebook.patch import patch

class ModuleImporter(NotebookImporter):
    pass

    def determine_package_name(self, path: Path) -> str:
        """Determine the corresponding importable name for a package directory given by
    a particular file path. Return `None` if path is not contained within `sys.path`.

    :param path: path to package
    :return:
    """
        for p in sys.path:
            if str(path) == p:
                continue
            try:
                relative_path = path.relative_to(p)
            except ValueError:
                continue
            return '.'.join(relative_path.parts)
        return None

    def install_hook(self):
        """Install notebook import hook

    Don't allow the user to specify a custom search path, because we also need this to
    interoperate with the default Python module importers which use sys.path

    :return:
    """
        cwd = Path.cwd()
        sys.path = [p for p in sys.path if Path(p).resolve() != cwd]
        super().install_hook()

    def update_namespace(self, namespace):
        cwd = Path.cwd()
        namespace.update({'__package__': self.determine_package_name(cwd), 'patch': patch})


""""""
import traitlets
import pathlib

class Path(traitlets.TraitType):
    """A pathlib.Path trait class"""

    def validate(self, obj, value):
        try:
            return pathlib.Path(value)
        except TypeError:
            self.error(obj, value)


"""# Commandline Application
"""
import sys
from traitlets import default
from traitlets.config import Application
from .build import LiteraryBuildApp
from .test import LiteraryTestApp

class LiteraryLauncher(Application):
    description = 'Work with literate notebooks'
    subcommands = {'build': (LiteraryBuildApp, 'Build a package from a series of notebooks'), 'test': (LiteraryTestApp, 'Run a series of notebook tests')}

    def start(self):
        """Perform the App's actions as configured"""
        super().start()
        if self.subapp is None:
            sub_commands = ', '.join(sorted(self.subcommands))
            sys.exit('Please supply at least one subcommand: {}'.format(sub_commands))
launch_new_instance = LiteraryLauncher.launch_instance


"""# Literary Application
"""
from copy import deepcopy
from importlib import import_module
from inspect import getmembers
from pathlib import Path
from traitlets import List, Unicode, default
from traitlets.config import Application, Configurable, catch_config_error
from ..core.config import find_project_config, load_project_config
from ..core.trait import Path as PathTrait

class LiteraryApp(Application):
    name = 'literary'
    description = 'A Literary application'
    aliases = {**Application.aliases, 'config-file': 'LiteraryApp.project_config_file'}
    project_config_file = PathTrait(help='Literary project configuration file').tag(config=True)
    classes = List()

    @default('classes')
    def _classes_default(self):
        modules = [import_module(f'..core.{n}', __package__) for n in ('exporter', 'package', 'preprocessor', 'project', 'test', 'transformers')]
        return [cls for m in modules for (_, cls) in getmembers(m) if isinstance(cls, type) and issubclass(cls, Configurable)]

    @default('project_config_file')
    def _project_config_file_default(self):
        return find_project_config(Path.cwd())

    @catch_config_error
    def initialize(self, argv=None):
        self.parse_command_line(argv)
        argv_config = deepcopy(self.config)
        self.load_app_config_file()
        self.update_config(argv_config)

    def load_app_config_file(self):
        config = load_project_config(self.project_config_file)
        self.update_config(config)


"""# Test Package Command
"""
import logging
from concurrent.futures import ProcessPoolExecutor
from pathlib import Path
from traitlets import Int, List, Unicode, default
from ..core.test import ProjectTester
from .app import LiteraryApp

class LiteraryTestApp(LiteraryApp):
    description = 'Test literary notebooks in parallel'
    aliases = {**LiteraryApp.aliases, 'ignore': 'ProjectTester.ignore_patterns', 'jobs': 'ProjectTester.jobs', 'packages': 'ProjectTester.packages_dir', 'extras': 'ProjectTester.extra_sources'}
    source = List(trait=Unicode(help='source directory or notebooks to run')).tag(config=True)
    jobs = Int(allow_none=True, default_value=None, help='number of parallel jobs to run').tag(config=True)
    ignore = List(help='glob pattern to ignore during recursion', trait=Unicode()).tag(config=True)

    def start(self):
        tester = ProjectTester(parent=self)
        tester.run()


"""# Build Package Command
"""
from pathlib import Path
from traitlets import List, Unicode, default
from ..core.package import PackageBuilder
from .app import LiteraryApp

class LiteraryBuildApp(LiteraryApp):
    description = 'Build a pure-Python package from a set of Jupyter notebooks'
    aliases = {**LiteraryApp.aliases, 'ignore': 'PackageBuilder.ignore_patterns', 'output': 'PackageBuilder.generated_dir', 'packages': 'PackageBuilder.packages_dir'}
    flags = {'clear': ({'PackageBuilder': {'clear_generated': True}}, 'Clear generated directory before building.')}

    def start(self):
        builder = PackageBuilder(parent=self)
        builder.build()


"""# Project Operations
Most Literary operations act upon the *project* - that is, the set of notebooks which are considered "packages". Here we define a `Configurable` which defines the configuration that refers to these notebooks
"""
from pathlib import Path
from traitlets import List, Unicode, default, observe, validate
from traitlets.config import Config, LoggingConfigurable
from .trait import Path as PathTrait

class ProjectOperator(LoggingConfigurable):
    project_path = PathTrait(help='Path to Literary project top-level directory').tag(config=True)
    packages_dir = Unicode('src', help='Path to Literary packages top-level directory').tag(config=True)

    @validate('project_path')
    def _validate_project_path(self, proposal):
        path = Path(proposal.value)
        if not path.exists():
            raise ValueError()
        return path

    def resolve_path(self, path):
        return self.project_path / path

    @property
    def packages_path(self) -> Path:
        return self.resolve_path(self.packages_dir)


"""# Package Generation
"""
import shutil
from pathlib import Path
import nbformat
from nbconvert import Exporter
from traitlets import Bool, Instance, List, Type, Unicode, default
from traitlets.config import Config, Configurable
from .exporter import LiteraryExporter
from .project import ProjectOperator
DEFAULT_IGNORE_PATTERNS = ('.ipynb_checkpoints', '__pycache__', '.*')

class PackageBuilder(ProjectOperator):
    """Project operator which builds a Literary package from a set of notebook directories."""
    exporter = Instance(Exporter)
    exporter_class = Type(LiteraryExporter).tag(config=True)
    generated_dir = Unicode('lib', help='Path to generated packages top-level directory').tag(config=True)
    ignore_patterns = List(Unicode(), help='List of patterns to ignore from source tree').tag(config=True)
    clear_generated = Bool(False, help='Clear generated directory before building, otherwise raise an Exception if non-empty.').tag(config=True)

    @default('exporter')
    def _exporter_default(self):
        return self.exporter_class(parent=self)

    @property
    def generated_path(self) -> Path:
        return self.resolve_path(self.generated_dir)

    @default('ignore_patterns')
    def _ignore_patterns_default(self):
        return list(DEFAULT_IGNORE_PATTERNS)

    def _build_package_component(self, source_dir_path: Path, dest_dir_path: Path):
        """Recursively build a pure-Python package from a source tree

    :param source_dir_path: path to current source directory
    :param dest_dir_path: path to current destination directory
    :return:
    """
        dest_dir_path.mkdir(parents=True, exist_ok=True)
        for path in source_dir_path.iterdir():
            if any((path.match(p) for p in self.ignore_patterns)):
                continue
            if path == self.generated_dir:
                continue
            relative_path = path.relative_to(source_dir_path)
            mirror_path = dest_dir_path / relative_path
            if path.match('*.ipynb'):
                (source, _) = self.exporter.from_notebook_node(nbformat.read(path, as_version=nbformat.NO_CONVERT))
                mirror_path.with_suffix('.py').write_text(source)
            elif path.is_dir():
                self._build_package_component(path, mirror_path)
            else:
                mirror_path.write_bytes(path.read_bytes())

    def _build_packages(self):
        """Build the packages contained in `packages_path`."""
        for p in self.packages_path.iterdir():
            self._build_package_component(p, self.generated_path / p.name)

    def _clear_generated_path(self):
        """Clear the contents of `generated_path`."""
        for p in self.generated_path.iterdir():
            if not self.clear_generated:
                raise ValueError('Generated directory is not empty, and `clear_generated` is not set.')
            if p.is_file():
                p.unlink()
            else:
                shutil.rmtree(p)

    def build(self):
        """Build a pure-Python package from a literary source tree."""
        self.generated_path.mkdir(parents=True, exist_ok=True)
        self._clear_generated_path()
        if not self.packages_path.exists():
            raise FileNotFoundError(f'Source path {self.packages_path!r} does not exist')
        self._build_packages()


"""# Standalone Notebook
"""
import sys
from pathlib import Path
from ..core.config import find_project_config, load_project_config
from .importer import NotebookImporter

def load_ipython_extension(ipython):
    """Load the import hook and setup the global state for the Literary extension.
    When IPython invokes this function, the determined package root path will be
    added to `sys.path`.

    :param ipython: IPython shell instance
    """
    config = load_project_config(find_project_config(Path.cwd()))
    importer = NotebookImporter(config=config)
    importer.install_hook()
    importer.update_namespace(ipython.user_ns)


""""""
import typing as tp
from functools import lru_cache
from pathlib import Path
from traitlets.config import Config, ConfigFileNotFound, JSONFileConfigLoader, PyFileConfigLoader
from .project import ProjectOperator
CONFIG_FILE_STEM = 'literary_config'

@lru_cache()
def find_project_config(path, *additional_paths) -> Path:
    """Load the configuration for the current Literary project.

    :param search_paths: starting search paths
    :return:
    """
    visited = set()
    for top_level_path in [path, *additional_paths]:
        for search_path in (top_level_path, *top_level_path.parents):
            if search_path in visited:
                break
            visited.add(search_path)
            for p in search_path.glob(f'{CONFIG_FILE_STEM}.*'):
                return p
    raise FileNotFoundError("Couldn't find config file")

def load_project_config(path: Path) -> Config:
    """Load a project configuration file

    :param path: configuration file path
    :return:
    """
    for loader_cls in (JSONFileConfigLoader, PyFileConfigLoader):
        loader = loader_cls(path.name, str(path.parent))
        try:
            config = loader.load_config()
            break
        except ConfigFileNotFound:
            continue
    else:
        raise ValueError(f'{path!r} was not a recognised config file')
    config.merge(Config({ProjectOperator.__name__: {'project_path': path.parent.absolute()}}))
    return config


"""# Notebook Finder
"""
import os
import pathlib
import sys
import traceback
import typing as tp
from importlib.machinery import FileFinder
from inspect import getclosurevars
from traitlets import Bool, Type
from ..core.exporter import LiteraryExporter
from ..core.project import ProjectOperator
T = tp.TypeVar('T')

def _get_loader_details(hook) -> tuple:
    """Return the loader_details for a given FileFinder closure

    :param hook: FileFinder closure
    :returns: loader_details tuple
    """
    try:
        namespace = getclosurevars(hook)
    except TypeError as err:
        raise ValueError from err
    try:
        return namespace.nonlocals['loader_details']
    except KeyError as err:
        raise ValueError from err

def _find_file_finder(path_hooks: list) -> tp.Tuple[int, tp.Any]:
    """Find the FileFinder closure in a list of path hooks

    :param path_hooks: path hooks
    :returns: index of hook and the hook itself
    """
    for (i, hook) in enumerate(path_hooks):
        try:
            _get_loader_details(hook)
        except ValueError:
            continue
        return (i, hook)
    raise ValueError

def _extend_file_finder(finder: T, *loader_details) -> T:
    """Extend an existing file finder with new loader details

    :param finder: existing FileFinder instance
    :param loader_details:
    :return:
    """
    return FileFinder.path_hook(*_get_loader_details(finder), *loader_details)

def inject_loaders(path_hooks: list, *loader_details):
    """Inject a set of loaders into a list of path hooks

    :param path_hooks: list of path hooks
    :param loader_details: FileFinder loader details
    :return:
    """
    (i, finder) = _find_file_finder(path_hooks)
    new_finder = _extend_file_finder(finder, *loader_details)
    path_hooks[i] = new_finder
    sys.path_importer_cache.clear()


"""# Notebook Loader
"""
import linecache
from importlib.machinery import SourcelessFileLoader
import nbformat

class NotebookLoader(SourcelessFileLoader):
    """Sourceless Jupyter Notebook loader"""

    def __init__(self, fullname: str, path: str, exporter):
        super().__init__(fullname, path)
        self._exporter = exporter

    def _update_linecache(self, path: str, source: str):
        linecache.cache[path] = (len(source), None, source.splitlines(keepends=True), path)

    def get_code(self, fullname: str):
        path = self.get_filename(fullname)
        body = self.get_transpiled_source(path)
        self._update_linecache(path, body)
        return compile(body, path, 'exec')

    def get_transpiled_source(self, path: str):
        nb = nbformat.read(path, as_version=nbformat.NO_CONVERT)
        (body, resources) = self._exporter.from_notebook_node(nb)
        return body


"""# Import Hook
"""
import os
import sys
import traceback
import typing as tp
from pathlib import Path
from nbconvert import Exporter
from traitlets import Bool, Instance, Type, default
from ..core.exporter import LiteraryExporter
from ..core.project import ProjectOperator
from .finder import inject_loaders
from .loader import NotebookLoader
from .patch import patch

class NotebookImporter(ProjectOperator):
    exporter = Instance(Exporter)
    exporter_class = Type(LiteraryExporter, help='Exporter class').tag(config=True)
    set_except_hook = Bool(True, help='Overwrite `sys.excepthook` to correctly display tracebacks').tag(config=True)

    @default('exporter')
    def _exporter_default(self):
        return self.exporter_class(parent=self)

    def install_hook(self):
        """Install notebook import hook

    Don't allow the user to specify a custom search path, because we also need this to
    interoperate with the default Python module importers which use sys.path

    :return:
    """
        sys.path.append(str(self.packages_path))
        exporter = self.exporter_class(parent=self)

        def create_notebook_loader(fullname, path):
            return NotebookLoader(fullname, path, exporter=exporter)
        inject_loaders(sys.path_hooks, (create_notebook_loader, ['.ipynb']))
        if self.set_except_hook:
            sys.excepthook = traceback.print_exception

    def update_namespace(self, namespace):
        namespace.update({'patch': patch})